import re

import pytest
from tabulate2 import tabulate

from ttrpg_dice import StatBlock, d, statblock


def test_empty():
    @statblock
    class Combat:
        WS = d(100)

    empty = Combat()
    assert empty.WS == 0


def test_isinstance_StatBlock():
    @statblock
    class Combat:
        WS = d(100)

    empty = Combat()
    assert isinstance(empty, StatBlock)


def test_instatiation():
    @statblock
    class Mixed:
        WS = d(100)
        Dex = d(100)

    fighter = Mixed(WS=41)

    assert fighter.WS == 41
    assert fighter.Dex == 0
    assert isinstance(fighter, StatBlock)
    assert type(fighter) is Mixed


def test_addition():
    @statblock
    class Combat:
        WS = d(100)

    fighter = Combat(WS=41)
    skilled = Combat(WS=10)

    knight = fighter + skilled

    assert knight.WS == 51
    assert isinstance(knight, Combat)


def test_instance_vars():
    @statblock
    class Combat:
        WS = d(100)

    fighter = Combat(WS=41)
    assert vars(fighter) == {"WS": 41}


def test__STATS():
    @statblock
    class Combat:
        WS = d(100)

    fighter = Combat(WS=41)

    assert fighter._STATS == {"WS": d(100)}  # noqa: SLF001


def test_maxed_out():
    @statblock
    class Combat:
        WS = d(100)

    fighter = Combat(WS=41)
    superhuman = Combat(WS=60)

    hero = fighter + superhuman

    assert hero.WS == 100


def test_union():
    @statblock
    class FullCombat:
        WS = d(100)
        BS = d(100)

    squire = FullCombat(WS=10, BS=10)
    thief = FullCombat(WS=20)

    runaway = squire | thief

    assert runaway.WS == 20
    assert runaway.BS == 10


def test_no_direct_instantiation():
    msg = re.escape("Cannot directly instantiate a StatBlock, please use the @statblock decorator instead.")
    with pytest.raises(TypeError, match=msg):
        _ = StatBlock()


def test_no_direct_instantiation_with_args():
    msg = re.escape("Cannot directly instantiate a StatBlock, please use the @statblock decorator instead.")
    with pytest.raises(TypeError, match=msg):
        _ = StatBlock(WS=7)


def test_subclass_has__STATS():
    @statblock
    class Combat:
        WS = d(100)

    class Human(Combat):
        WS = 33

    albert = Human()
    assert albert._STATS == {"WS": d(100)}  # noqa: SLF001


def test_subclass_stat():
    @statblock
    class Combat:
        WS = d(100)

    class Human(Combat):
        WS = 33

    albert = Human()
    assert albert.WS == 33


def test_subclass_partial():
    @statblock
    class FullCombat:
        WS = d(100)
        BS = d(100)

    class Human(FullCombat):
        WS = 33

    albert = Human()
    assert vars(albert) == {"WS": 33, "BS": 0}

def test_subclass_custom_init():

    @statblock
    class FullCombat:
        WS = d(100)
        BS = d(100)

        def _pre_init_(self, /, name, **kwargs):
            self.name = name
            return kwargs
    
    class Human(FullCombat):
        WS = 33

    albert = Human(name="Albert")

    assert albert.name == "Albert"


def test_kw_only():
    @statblock
    class FullCombat:
        WS = d(100)
        BS = d(100)

    with pytest.raises(TypeError):
        _ = FullCombat(45)

def test_invalid_stat():
    @statblock
    class FullCombat:
        WS = d(100)
        BS = d(100)

    msg = "Invalid stat. FullCombat does not contain `M`."
    with pytest.raises(AttributeError, match = msg):
        _ = FullCombat(WS=33, M = 4)

def test_str():
    @statblock
    class FullCombat:
        WS = d(100)
        BS = d(100)

    class Human(FullCombat):
        WS = 33

    albert = Human()

    assert str(albert) == "Human FullCombat StatBlock"

def test_repr():
    @statblock
    class FullCombat:
        WS = d(100)
        BS = d(100)

    class Human(FullCombat):
        WS = 33

    albert = Human()

    assert repr(albert) == "Human FullCombat StatBlock(WS: d100 = 33, BS: d100 = 0)"


def test_subscripting():
    @statblock
    class FullCombat:
        WS = d(100)
        BS = d(100)

    class Human(FullCombat):
        WS = 33

    albert = Human()

    assert albert["WS"] == 33

def test_subscripting_invalidstat():
    @statblock
    class FullCombat:
        WS = d(100)
        BS = d(100)

    class Human(FullCombat):
        WS = 33

    albert = Human()

    msg = "Unknown stat 'M'"
    with pytest.raises(KeyError, match=msg):
        assert albert["M"] == 33

def test_mapping():
    @statblock
    class FullCombat:
        WS = d(100)
        BS = d(100)

    class Human(FullCombat):
        WS = 33

    albert = Human()

    assert list(albert.keys()) == ["WS", "BS"]
    assert list(albert.values()) == [33, 0]
    assert list(albert.items()) == [("WS", 33), ("BS", 0)]

def test_table():
    @statblock
    class FullCombat:
        WS = d(100)
        BS = d(100)

    class Human(FullCombat):
        WS = 33

    albert = Human()

    assert albert.as_table() == tabulate([[33,0]], headers=["WS","BS"], tablefmt="github")

# TODO: type-hinting instances (https://docs.python.org/3/library/typing.html#typing.get_type_hints)
# TODO: Handle `@statblock()` usage
# TODO: Maths where blocks have different stats
# TODO: Make stats immutable and enable hash (allowing usage as a dict key)



"""Create statblocks easily."""

from __future__ import annotations

from collections.abc import Iterator, Mapping
from typing import TYPE_CHECKING

from tabulate2 import tabulate

from .dice import Dice

if TYPE_CHECKING:
    from contextlib import suppress

    with suppress(ImportError):
        from typing import ClassVar, Self


class StatBlock(Mapping):
    """A TTRPG StatBlock, acts as a mapping of stats."""

    _STATS: ClassVar[dict[str, Dice]]

    def __init__(self, *_args, **_kwargs) -> None:  # noqa: ANN002, ANN003
        """Raises TypeError - this method is replaced by `_init_` by the decorator."""
        if type(self) is StatBlock:
            msg = "Cannot directly instantiate a StatBlock, please use the @statblock decorator instead."
            raise TypeError(msg)
        
    def _pre_init_(self, /, **stats: int) -> dict[str, int | Dice]:
        """
        Subclasses can override this to perform actions at the start of __init__.
        
        Be sure to pop any used kwargs and return the remaining ones.
        """
        return stats
    
    def _init_(self, /, **stats: int | Dice) -> None:
        """Initialise a StatBlock with some, or all stats given."""
        stats = self._pre_init_(**stats)
        for stat in self._STATS:
            val = stats.pop(stat, vars(type(self)).get(stat, 0))
            setattr(self, stat, val)
        if stats:
            remaining_stats = ", ".join(f"`{stat}`" for stat in stats)
            msg = f"Invalid stat. {type(self).__name__} does not contain {remaining_stats}."
            raise AttributeError(msg)

    def __add__(self, other: Self) -> Self:
        """Adds each stat, raises AttributeError if stat missing in `other`."""
        newstats = {
            stat: min(getattr(self, stat) + getattr(other, stat), len(self._STATS[stat])) for stat in self._STATS
        }
        return type(self)(**newstats)

    def __or__(self, other: Self) -> Self:
        """Merge stats, keeping the highest."""
        newstats = {stat: max(getattr(self, stat), getattr(other, stat)) for stat in self._STATS}
        return type(self)(**newstats)

    def __getitem__(self, stat: str) -> int | Dice:
        """Get a specific stat by subscripting."""
        if stat in self._STATS:
            return getattr(self, stat)
        msg = f"Unknown stat '{stat}'"
        raise KeyError(msg)
    
    def __len__(self) -> int:
        """Number of stats."""
        return len(self._STATS)
    
    def __iter__(self) -> Iterator:
        """Iterate over stats."""
        return iter(self._STATS)
    
    def __str__(self) -> str:
        """A description of the Statblock type e.g. 'Human Warhammer StatBlock'."""
        cls = type(self)
        bases = cls.mro()
        statblock_index = bases.index(StatBlock)
        return " ".join(base.__name__ for base in bases[:statblock_index+1])

    def __repr__(self) -> str:
        """Statblock type as per `str` plus the stats and their challenge rolls."""
        return (
            str(self)
            + "("
            + ", ".join(f"{statname}: {roll} = {self[statname]}" for statname, roll in self._STATS.items())
            + ")"
        )
    
    def as_table(self) -> str:
        """Render the StatBlock as a github markdown table."""
        return tabulate([[*self.values()]], headers=self.keys(), tablefmt="github")

    def _repr_markdown_(self) -> str:
        """For IPython notebooks - L3 header and stat table."""
        return f"### {self}\n{self.as_table()}"

def statblock(cls: type) -> StatBlock:
    """Create a StatBlock with the given fields."""
    stats = {statname: roll for statname, roll in vars(cls).items() if isinstance(roll, Dice)}
    _interimclass: type = type(
        cls.__name__,
        (StatBlock,),
        {attr: 0 if attr in stats else val for attr, val in vars(cls).items() if attr != "__dict__"},
    )
    _interimclass.__annotations__ = dict.fromkeys(stats, int | Dice)
    _interimclass._STATS = stats  # noqa: SLF001
    _interimclass.__init__ = StatBlock._init_
    return _interimclass



from __future__ import annotations

from dataclasses import dataclass
from math import isclose
from operator import indexOf
from typing import Any

import pytest  # noqa: F401, RUF100

from ttrpg_dice import d


@dataclass
class DiceTest:
    dice_expr: str
    id: str
    description: str
    repr_: str
    contents: dict
    hashed: int
    probabilities: list
    weighted: bool
    faces: int


# fmt: off
DiceTests = [
    DiceTest(
        dice_expr="d(4)",
        description="d4",
        repr_="Dice: d4 ({4: 1})",
        contents={4: 1},
        hashed=hash(((4,1),)),
        probabilities=[0.25, 0.25, 0.25, 0.25],
        weighted=False,
        faces=4,
        id="d4",
    ),
    DiceTest(
        dice_expr="d(100)",
        description="d100",
        repr_="Dice: d100 ({100: 1})",
        contents={100: 1},
        hashed=hash(((100,1),)),
        probabilities=[0.01] * 100,
        weighted=False,
        faces=100,
        id="d100",
    ),
    DiceTest(
        dice_expr="2 * d(4)",
        description="2d4",
        repr_="Dice: 2d4 ({4: 2})",
        contents={4: 2},
        hashed=hash(((4,2),)),
        probabilities=[0, 0.0625, 0.125, 0.1875, 0.25, 0.1875, 0.125, 0.0625],
        weighted=True,
        faces=8,
        id="multiplication",
    ),
    DiceTest(
        dice_expr="d(2) + d(4)",
        description="d2 + d4",
        repr_="Dice: d2 + d4 ({2: 1, 4: 1})",
        contents={2: 1, 4: 1},
        hashed=hash(((2,1),(4,1))),
        probabilities=[0, 0.125, 0.25, 0.25, 0.25, 0.125],
        weighted=True,
        faces=6,
        id="addition",
    ),
    DiceTest(
        dice_expr="d(4) + 2",
        description="d4 + 2",
        repr_="Dice: d4 + 2 ({1: 2, 4: 1})",
        contents={1: 2, 4: 1},
        hashed=hash(((1,2),(4,1))),
        probabilities=[0, 0, 0.25, 0.25, 0.25, 0.25],
        weighted=True,
        faces=6,
        id="add constant",
    ),
    DiceTest(
        dice_expr="d(4) + 1",
        description="d4 + 1",
        repr_="Dice: d4 + 1 ({1: 1, 4: 1})",
        contents={1: 1, 4: 1},
        hashed=hash(((1,1),(4,1))),
        probabilities=[0, 0.25, 0.25, 0.25, 0.25],
        weighted=True,
        faces=5,
        id="add 1",
    ),
    DiceTest(
        dice_expr="d(6) + d(4)",
        description="d4 + d6",
        repr_="Dice: d4 + d6 ({4: 1, 6: 1})",
        contents={4: 1, 6: 1},
        hashed=hash(((4,1),(6,1))),
        probabilities=[
            0,
            0.0416666666667,
            0.0833333333333,
            0.125,
            0.166666666667,
            0.166666666667,
            0.166666666667,
            0.125,
            0.0833333333333,
            0.0416666666667,
        ],
        weighted=True,
        faces=10,
        id="unsorted addition: two dice",
    ),
    DiceTest(
        dice_expr="(2 * d(2)) + d(4)",
        description="2d2 + d4",
        repr_="Dice: 2d2 + d4 ({2: 2, 4: 1})",
        contents={2: 2, 4: 1},
        hashed=hash(((2,2),(4,1))),
        probabilities=[0, 0, 0.0625, 0.1875, 0.25, 0.25, 0.1875, 0.0625],
        weighted=True,
        faces=8,
        id="addition: complex dice",
    ),
    DiceTest(
        dice_expr="d(4) + (2 * d(3)) + 1",
        description="2d3 + d4 + 1",
        repr_="Dice: 2d3 + d4 + 1 ({1: 1, 3: 2, 4: 1})",
        contents={1: 1, 3: 2, 4: 1},
        hashed=hash(((1,1),(3,2),(4,1))),
        probabilities=[
            0,
            0,
            0,
            0.0277777777778,
            0.0833333333333,
            0.166666666667,
            0.222222222222,
            0.222222222222,
            0.166666666667,
            0.0833333333333,
            0.0277777777778,
        ],
        weighted=True,
        faces=11,
        id="combined arithmetic",
    ),
    DiceTest(
        dice_expr="d(8) + (2 * d(8))",
        description="3d8",
        repr_="Dice: 3d8 ({8: 3})",
        contents={8: 3},
        hashed=hash(((8,3),)),
        probabilities=[
            0,
            0,
            0.001953125,
            0.005859375,
            0.01171875,
            0.01953125,
            0.029296875,
            0.041015625,
            0.0546875,
            0.0703125,
            0.08203125,
            0.08984375,
            0.09375,
            0.09375,
            0.08984375,
            0.08203125,
            0.0703125,
            0.0546875,
            0.041015625,
            0.029296875,
            0.01953125,
            0.01171875,
            0.005859375,
            0.001953125,
        ],
        weighted=True,
        faces=24,
        id="add similar dice",
    ),
]
# fmt: on


@pytest.mark.parametrize(
    ["dietype", "probabilities"],
    [(pytest.param(tc.dice_expr, tc.probabilities, id=tc.id)) for tc in DiceTests],
    indirect=["dietype"],
)
def test_probabilities(dietype, probabilities):
    check = [isclose(p, e) for p, e in zip(list(dietype), probabilities)]  # noqa: B905
    try:
        mismatch = indexOf(check, False)  # noqa: FBT003
        msg = f"First mismatch p({mismatch}) is {list(dietype)[mismatch]} should be {probabilities[mismatch]}"
    except ValueError:
        pass
    assert all(check), msg


@pytest.mark.parametrize(
    ["dietype", "description"],
    [pytest.param(tc.dice_expr, tc.description, id=tc.id) for tc in DiceTests],
    indirect=["dietype"],
)
def test_str(dietype, description):
    assert str(dietype) == description


@pytest.mark.parametrize(
    ["dietype", "repr_"],
    [(pytest.param(tc.dice_expr, tc.repr_, id=tc.id)) for tc in DiceTests],
    indirect=["dietype"],
)
def test_repr(dietype, repr_):
    assert repr(dietype) == repr_


@pytest.mark.parametrize(
    ["dietype", "contents"],
    [(pytest.param(tc.dice_expr, tc.contents, id=tc.id)) for tc in DiceTests],
    indirect=["dietype"],
)
def test_contents(dietype, contents):
    assert dietype.contents == contents


@pytest.mark.parametrize(
    ["dietype", "hashed"],
    [(pytest.param(tc.dice_expr, tc.hashed, id=tc.id)) for tc in DiceTests],
    indirect=["dietype"],
)
def test_hash(dietype, hashed):
    assert hash(dietype) == hashed


@pytest.mark.parametrize(
    ["dietype", "contents"],
    [(pytest.param(tc.dice_expr, tc.contents, id=tc.id)) for tc in DiceTests],
    indirect=["dietype"],
)
def test_fromcontents(dietype: d, contents: dict):
    assert d.from_contents(contents) == dietype

@pytest.mark.parametrize(
    ["description", "dietype"],
    [pytest.param(tc.description, tc.dice_expr, id=tc.id) for tc in DiceTests],
    indirect=["dietype"],
)
def test_from_str(description: str, dietype: d):
    assert d.from_str(description) == dietype

@pytest.mark.parametrize(
    ["dietype", "weighted"],
    [(pytest.param(tc.dice_expr, tc.weighted, id=tc.id)) for tc in DiceTests],
    indirect=["dietype"],
)
def test_weighted(dietype: d, weighted: bool):  # noqa: FBT001
    assert dietype.weighted == weighted


@pytest.mark.parametrize(
    ["dietype", "probabilities"],
    [(pytest.param(tc.dice_expr, tc.probabilities, id=tc.id)) for tc in DiceTests],
    indirect=["dietype"],
)
def test_first_probability(dietype, probabilities):
    assert isclose(dietype[1], probabilities[0])


@pytest.mark.parametrize(
    ["dietype", "faces"],
    [(pytest.param(tc.dice_expr, tc.faces, id=tc.id)) for tc in DiceTests],
    indirect=["dietype"],
)
def test_faces(dietype: d, faces: int):
    assert len(dietype) == faces


@dataclass
class InvalidDiceTestCase:
    errortype: Exception
    errormsg: str
    id: str
    contents: dict | None = None
    faces: Any | None = None


# fmt: off
invalid_faces_cases = [
    InvalidDiceTestCase(
        faces=0,
        errortype=ValueError,
        errormsg="Number of faces must be a positive integer, not 0",
        id="zero",
    ),
    InvalidDiceTestCase(
        faces=-1,
        errortype=ValueError,
        errormsg="Number of faces must be a positive integer, not -1",
        id="negative",
    ),
    InvalidDiceTestCase(
        faces="foo",
        errortype=TypeError,
        errormsg="Number of faces must be a positive integer, not str",
        id="str",
    ),
    InvalidDiceTestCase(
        faces=1.5,
        errortype=TypeError,
        errormsg="Number of faces must be a positive integer, not float",
        id="float",
    ),
]

invalid_contents_cases = [
    InvalidDiceTestCase(
        contents={1: "2"},
        errortype=TypeError,
        errormsg="Number of Dice must be a positive integer, not str",
        id="numdice: str",
    ),
    InvalidDiceTestCase(
        contents={1: 2.0},
        errortype=TypeError,
        errormsg="Number of Dice must be a positive integer, not float",
        id="numdice: float",
    ),
    InvalidDiceTestCase(
        contents={1: -1},
        errortype=ValueError,
        errormsg="Number of Dice must be a positive integer, not -1",
        id="numdice: negative",
    ),
    InvalidDiceTestCase(
        contents={"foo": 1},
        errortype=TypeError,
        errormsg="Number of faces must be a positive integer, not str",
        id="faces: str",
    ),
    InvalidDiceTestCase(
        contents={1.5: 1},
        errortype=TypeError,
        errormsg="Number of faces must be a positive integer, not float",
        id="faces: float",
    ),
    InvalidDiceTestCase(
        contents={0: 1},
        errortype=ValueError,
        errormsg="Number of faces must be a positive integer, not 0",
        id="faces: zero",
    ),
    InvalidDiceTestCase(
        contents={-1: 1},
        errortype=ValueError,
        errormsg="Number of faces must be a positive integer, not -1",
        id="faces: negative",
    ),
    InvalidDiceTestCase(
        contents={4: 3, 2: 3.2, 1: "2"},
        errortype=TypeError,
        errormsg="Number of Dice must be a positive integer, not str, float",
        id="numdice: partially valid types",
    ),
    InvalidDiceTestCase(
        contents={4: 3, 2: -1, 1: -2},
        errortype=ValueError,
        errormsg="Number of Dice must be a positive integer, not -2, -1",
        id="numdice: partially valid values",
    ),
    InvalidDiceTestCase(
        contents={5: 2, "-1": 1, 1: 2, 3.2: 4},
        errortype=TypeError,
        errormsg="Number of faces must be a positive integer, not float, str",
        id="faces: partially valid types",
    ),
    InvalidDiceTestCase(
        contents={5: 2, -1: 1, 1: 2, -2: 3},
        errortype=ValueError,
        errormsg="Number of faces must be a positive integer, not -2, -1",
        id="faces: partially valid values",
    ),
]
# fmt: on


@pytest.mark.parametrize(
    ["faces", "errortype", "errormsg"],
    [pytest.param(tc.faces, tc.errortype, tc.errormsg, id=tc.id) for tc in invalid_faces_cases],
)
def test_invalid_die(faces, errortype, errormsg):
    with pytest.raises(errortype, match=errormsg):
        d(faces)


@pytest.mark.parametrize(
    ["contents", "errortype", "errormsg"],
    [pytest.param(tc.contents, tc.errortype, tc.errormsg, id=tc.id) for tc in invalid_contents_cases],
)
def test_invalid_from_contents(contents, errortype, errormsg):
    with pytest.raises(errortype, match=errormsg):
        d.from_contents(contents)



import pytest

from ttrpg_dice import d


@pytest.fixture
def dietype(request: str) -> d:
    """
    Instatiate dice from string using `d` for `Dice`, at execution time.

    Intended for use in parameterised tests requiring Dice allowing for any instantiation errors to occur during
    execution and not test collection.

    Usage:
    ```
    @pytest.mark.parametrize(
        ["dietype", "description"],
        [
            ("d(4)", "d4"),
        ],
        indirect=["dietype"],
    )
    """
    return eval(request.param)  # noqa: S307



from __future__ import annotations

import re
from dataclasses import dataclass
from math import isclose
from operator import indexOf

import pytest  # noqa: F401, RUF100


@dataclass
class SliceTest:
    id: str
    index: slice | int
    dice_expr: str
    probabilities: list | None = None
    errortype: Exception | None = None
    errormsg: str | None = None


# fmt: off
SliceTests = [
    SliceTest(
        dice_expr="2 * d(4)",
        index=slice(None, None),
        probabilities=[0, 0.0625, 0.125, 0.1875, 0.25, 0.1875, 0.125, 0.0625],
        id="full slice",
    ),
    SliceTest(
        dice_expr="2 * d(4)",
        index=slice(2, 5),
        probabilities=[0.0625, 0.125, 0.1875],
        id="middle section",
    ),
    SliceTest(
        dice_expr="2 * d(4)",
        index=slice(None, 5),
        probabilities=[0, 0.0625, 0.125, 0.1875],
        id="from start",
    ),
    SliceTest(
        dice_expr="2 * d(4)",
        index=slice(3, None),
        probabilities=[0.125, 0.1875, 0.25, 0.1875, 0.125, 0.0625],
        id="to end",
    ),
    SliceTest(
        dice_expr="2 * d(4)",
        index=slice(None, None, -1),
        probabilities=[0.0625, 0.125, 0.1875, 0.25, 0.1875, 0.125, 0.0625, 0],
        id="reverse full slice",
    ),
    SliceTest(
        dice_expr="2 * d(4)",
        index=slice(-1, 0, -1),
        probabilities=[0.0625, 0.125, 0.1875, 0.25, 0.1875, 0.125, 0.0625, 0],
        id="reverse full slice to zero",
    ),
    SliceTest(
        dice_expr="2 * d(4)",
        index=slice(None, 1, -1),
        probabilities=[0.0625, 0.125, 0.1875, 0.25, 0.1875, 0.125, 0.0625, 0],
        id="reverse slice to 2",
    ),
    SliceTest(
        dice_expr="2 * d(4)",
        index=slice(7, 4, -1),
        probabilities=[0.125, 0.1875, 0.25],
        id="reverse middle section",
    ),
    SliceTest(
        dice_expr="2 * d(4)",
        index=slice(None, 3, -1),
        probabilities=[0.0625, 0.125, 0.1875, 0.25, 0.1875],
        id="reverse from end",
    ),
    SliceTest(
        dice_expr="2 * d(4)",
        index=slice(5, None, -1),
        probabilities=[0.25, 0.1875, 0.125, 0.0625, 0],
        id="reverse to start",
    ),
    SliceTest(
        dice_expr="2 * d(4)",
        index=slice(2, None, 2),
        probabilities=[0.0625, 0.1875, 0.1875, 0.0625],
        id="explicit evens",
    ),
    SliceTest(
        dice_expr="2 * d(4)",
        index=slice(None, None, 2),
        probabilities=[0.0625, 0.1875, 0.1875, 0.0625],
        id="implicit evens",
    ),
    SliceTest(
        dice_expr="2 * d(4)",
        index=slice(1, None, 2),
        probabilities=[0, 0.125, 0.25, 0.125],
        id="odds",
    ),
]
# fmt: on


@pytest.mark.parametrize(
    ["dietype", "sides", "probabilities"],
    [pytest.param(tc.dice_expr, tc.index, tc.probabilities, id=tc.id) for tc in SliceTests],
    indirect=["dietype"],
)
def test_slicing(dietype, sides, probabilities):
    check = [isclose(p, e) for p, e in zip(dietype[sides], probabilities)]  # noqa: B905
    try:
        mismatch = indexOf(check, False)  # noqa: FBT003
        msg = f"First mismatch p({mismatch}) is {list(dietype)[mismatch]} should be {probabilities[mismatch]}"
    except ValueError:
        pass
    assert all(check), msg


# fmt: off
IndexTests = [
    SliceTest(
        id="1",
        dice_expr="2 * d(4)",
        index=1,
        probabilities=0,
    ),
    SliceTest(
        id="2",
        dice_expr="2 * d(4)",
        index=2,
        probabilities=0.0625,
    ),
    SliceTest(
        id="-1",
        dice_expr="2 * d(4)",
        index=-1,
        probabilities=0.0625,
    ),
    SliceTest(
        id="-2",
        dice_expr="2 * d(4)",
        index=-2,
        probabilities=0.125,
    ),
    SliceTest(
        id="8 of 8",
        dice_expr="2 * d(4)",
        index=8,
        probabilities=0.0625,
    ),
    SliceTest(
        id="-8 of 8",
        dice_expr="2 * d(4)",
        index=-8,
        probabilities=0,
    ),
]
# fmt: on


@pytest.mark.parametrize(
    ["dietype", "side", "probability"],
    [pytest.param(tc.dice_expr, tc.index, tc.probabilities, id=tc.id) for tc in IndexTests],
    indirect=["dietype"],
)
def test_indexing(dietype, side, probability):
    assert dietype[side] == pytest.approx(probability)


# fmt: off
InvalidIndexTests = [
    SliceTest(
        id="zero",
        index=0,
        dice_expr="d(10)",
        errortype=IndexError,
        errormsg="Invalid side: This Dice has sides numbered 1 to 10.",
    ),
    SliceTest(
        id="too high",
        index=11,
        dice_expr="d(10)",
        errortype=IndexError,
        errormsg="Invalid side: This Dice has sides numbered 1 to 10.",
    ),
    SliceTest(
        id="negative results in p(0)",
        index=-11,
        dice_expr="d(10)",
        errortype=IndexError,
        errormsg="Invalid side: This Dice has sides numbered 1 to 10.",
    ),
    SliceTest(
        id="too big negative",
        index=-12,
        dice_expr="d(10)",
        errortype=IndexError,
        errormsg="Invalid side: This Dice has sides numbered 1 to 10.",
    ),
    SliceTest(
        id="number as string",
        index="3",
        dice_expr="d(10)",
        errortype=TypeError,
        errormsg="Cannot index 'Dice' with 'str'.",
    ),
    SliceTest(
        id="slice from zero",
        index=slice(0, None, None),
        dice_expr="d(10)",
        errortype=IndexError,
        errormsg="Invalid side: This Dice has sides numbered 1 to 10.",
    ),
]
# fmt: on


@pytest.mark.parametrize(
    ["dietype", "index", "errortype", "errormsg"],
    [pytest.param(tc.dice_expr, tc.index, tc.errortype, tc.errormsg, id=tc.id) for tc in InvalidIndexTests],
    indirect=["dietype"],
)
def test_invalid_index(dietype, index, errortype, errormsg):
    msg = re.escape(errormsg)
    with pytest.raises(errortype, match=msg):
        dietype[index]



"""A Dice class."""

from __future__ import annotations

from collections import defaultdict, deque
from itertools import product, repeat
from typing import TYPE_CHECKING, SupportsInt

if TYPE_CHECKING:
    from collections.abc import Generator, Iterator

try:
    from typing import Self
except ImportError:
    from typing_extensions import Self  # required for Python3.10 and below


class Dice:
    """A Dice class."""

    class _Contents(defaultdict):
        """Dice Contents are immutable after creation and return `0` if queried for a missing entry."""

        def __init__(self, *args, **kwargs) -> None:  # noqa: ANN002, ANN003
            tempdict = dict(*args, **kwargs)
            super().__init__(int, {faces: numdice for faces, numdice in tempdict.items() if numdice != 0})
            self._validate()

        def _validate(self) -> None:
            """
            `Dice._probabilities()` is called lazily and will be very hard to debug if contents are not valid.

            - `range(1,faces)` requires positive `int`
            - `repeat(...,numdice)` requires postive `int`
            """
            int_faces = {faces: isinstance(faces, int) for faces in self.keys()}
            if not all(int_faces.values()):
                invalid = ", ".join(sorted(type(faces).__name__ for faces, isint in int_faces.items() if not isint))
                msg = f"Number of faces must be a positive integer, not {invalid}"
                raise TypeError(msg)

            positive_faces = {faces: faces > 0 for faces in self.keys()}
            if not all(positive_faces.values()):
                invalid = ", ".join(
                    str(faces) for faces, ispositive in sorted(positive_faces.items()) if not ispositive
                )
                msg = f"Number of faces must be a positive integer, not {invalid}"
                raise ValueError(msg)

            int_numdice = {faces: isinstance(numdice, int) for faces, numdice in self.items()}
            if not all(int_numdice.values()):
                invalid = ", ".join(
                    type(self[faces]).__name__ for faces, isint in sorted(int_numdice.items()) if not isint
                )
                msg = f"Number of Dice must be a positive integer, not {invalid}"
                raise TypeError(msg)

            positive_numdice = {faces: numdice > 0 for faces, numdice in self.items()}
            if not all(positive_numdice.values()):
                invalid = ", ".join(
                    str(self[faces]) for faces, ispositive in sorted(positive_numdice.items()) if not ispositive
                )
                msg = f"Number of Dice must be a positive integer, not {invalid}"
                raise ValueError(msg)

        def __setitem__(self, key: int, value: int):  # noqa: ANN204
            msg = "Dice contents cannot be changed"
            raise TypeError(msg)

        def __missing__(self, key):  # noqa: ANN001, ANN204
            """Does not set new entry if Contents are frozen."""
            return self.default_factory()

        def __hash__(self) -> int:
            """Hash contents as tuple of sorted (key, value) tuples."""
            return hash(tuple(sorted(self.items())))

    def __init__(self, faces: int) -> None:
        """Build a die with `faces` sides."""
        self.contents = self._Contents({faces: 1})

    @property
    def _probabilities(self) -> list[float | None]:
        """
        Use Dice[index] to get the probability(-ies) of a given (set of) roll(s) NOT _probabilities.

        If for some reason you MUST access the underlying probabilities list use this property and not
        Dice._probabilitycache which is created lazily on the first call to this method.

        Returns a list of P(result) with _probabilities[0] = `None`.
        """
        try:
            return self._probabilitycache  # pytype: disable=attribute-error
        except AttributeError:
            all_possible_rolls = [sum(r) for r in product(*self._individual_dice_rolls())]
            ways_to_roll = {roll: all_possible_rolls.count(roll) for roll in range(1, max(all_possible_rolls) + 1)}
            number_possible_rolls = len(all_possible_rolls)
            self._probabilitycache = [None] + [n / number_possible_rolls for _, n in sorted(ways_to_roll.items())]
            return self._probabilitycache

    @_probabilities.setter
    def _probabilities(self, _: None) -> None:
        msg = "You cannot change a Dice's probabilities, create a new Dice instead."
        raise AttributeError(msg)

    @property
    def weighted(self) -> bool:
        """Is this Dice weighted, or are all results equally likely?"""
        return min(self) != max(self)

    def __iter__(self) -> Iterator:
        """Iterating over a Dice yields the probabilities starting with P(1)."""
        yield from self._probabilities[1:]

    def __getitem__(self, index: int | slice) -> float | list[float] | None:
        """
        Get the probability of a specific result, or a list of probabilities in the case of a slice.

        This handles the fact that dice faces are numbered from 1, not 0. Slicing with a step > 1 will
        return the probabilities of results beginning with the given step, not beginning with 1.

        Example:
            ```
            >>> from ttrpg_dice import d
            >>> dice = 2 * d(2)
            >>> list(dice)
            [0.0, 0.25, 0.5, 0.25]

            >>> dice[:]
            [0.0, 0.25, 0.5, 0.25]

            >>> dice[1]
            0.0

            >>> dice[::2] # evens
            [0.25, 0.25]

            >>> dice[1::2] # odds
            [0.0, 0.5]
            ```
        """
        if index == 0 or index == -(len(self) + 1):
            raise DiceIndexError(self)

        try:
            # pytype: disable=attribute-error
            if index.step is None or index.step > 0:  # Positive step
                if index.start is None:
                    index = slice(1 if index.step is None else index.step, index.stop, index.step)
                elif index.start == 0:  # To avoid possible confusion by slicing [0:]
                    raise DiceIndexError(self)
            else:  # Negative Step  # noqa: PLR5501
                if index.stop is None:
                    index = slice(index.start, 0, index.step)
            # pytype: enable=attribute-error
        except AttributeError:
            pass

        try:
            return self._probabilities[index]
        except TypeError as e:
            msg = f"Cannot index '{type(self).__name__}' with '{type(index).__name__}'."
            raise TypeError(msg) from e
        except IndexError as e:
            raise DiceIndexError(self) from e

    def __eq__(self, value: object) -> bool:
        """Dice are equal if they give the same probabilities, even with different contents."""
        try:
            return self._probabilities == value._probabilities  # pytype: disable=attribute-error
        except AttributeError:
            return False

    def __hash__(self) -> int:
        """Use contents for hashing - but NOT equality."""
        return hash(self.contents)

    def __len__(self) -> int:
        """Number of faces."""
        return len(self._probabilities) - 1

    def __str__(self) -> str:
        """The type of Dice in NdX notation."""
        sortedcontents = deque(sorted(self.contents.items()))

        # Place any constant at the end ("d4 + 2" not "2 + d4")
        if sortedcontents[0][0] == 1:
            sortedcontents.rotate(-1)

        def _ndx(n: int, x: int) -> str:
            if x == 1:
                return str(n)
            return f"{n if n > 1 else ''}d{x}"

        return " + ".join(_ndx(n, x) for x, n in sortedcontents)

    def __repr__(self) -> str:
        """Classname: ndX (contents)."""
        contents = ", ".join([f"{d}: {n}" for d, n in self.contents.items()])
        return f"{type(self).__name__}: {self} ({{{contents}}})"

    # =================
    # Block of stuff that returns Self ... pytype doesn't like this while we have Python3.10 and below
    # pytype: disable=invalid-annotation

    def __rmul__(self, other: SupportsInt) -> Self:
        """2 * Dice(4) returns a Dice with probabilities for 2d4."""
        other = self._int(other, "multiply", "by")
        return self.from_contents({f: n * other for f, n in self.contents.items()})

    def __add__(self, other: Self | SupportsInt) -> Self:
        """Adding two Dice to gives the combined roll."""
        try:
            othercontents = other.contents  # pytype: disable=attribute-error
        except AttributeError:
            othercontents = defaultdict(int, {1: self._int(other, "add", "and")})

        contents = {
            face: self.contents[face] + othercontents[face] for face in set(self.contents.keys() | othercontents.keys())
        }

        return self.from_contents(contents)

    @classmethod
    def from_contents(cls, contents: dict) -> Self:
        """Create a new die from a dict of contents."""
        die = cls.__new__(cls)
        die.contents = cls._Contents(contents)
        return die
    
    @classmethod
    def from_str(cls, description: str) -> Self:
        """Create a new die from ndX notation."""
        dice = description.split("+")
        contents = {
            int(x) if x else 1: int(n) if n else 1
            for die in dice
            for n, d, x in [die.strip().partition("d")]
        }
        return cls.from_contents(contents)

    # pytype: enable=invalid-annotation
    # END Block of stuff that returns Self ... pytype doesn't like this while we have Python3.10 and below
    # =================

    def _individual_dice_rolls(self) -> Generator[list, None, None]:
        """
        Yields a series of lists, each containing the valid faces of the individual dice contained within this `Dice`.

        Examples:
            For a simple d4:
            ```
            >>> list(Dice(4)._individual_dice_rolls())
            [[1, 2, 3, 4]]
            ```

            For 2d4 + d6 + 2:
            ```
            >>> from ttrpg_dice import d
            >>> dice = (2*d(4)) + d(6) + 2
            >>> list(dice._individual_dice_rolls())
            [[1], [1], [1, 2, 3, 4], [1, 2, 3, 4], [1, 2, 3, 4, 5, 6]]
            ```
        """
        for faces, numdice in self.contents.items():
            yield from repeat(list(range(1, faces + 1)), numdice)

    @classmethod
    def _int(cls, other: SupportsInt, action: str, conjunction: str) -> int:
        """Attempts to convert `other` to an int for use in arithmetic magic methods."""
        try:
            other = int(other)
        except TypeError as e:
            msg = f"Cannot {action} '{type(other).__name__}' {conjunction} '{cls.__name__}'"
            raise TypeError(msg) from e
        except ValueError as e:
            msg = f"Cannot {action} '{other}' {conjunction} '{cls.__name__}'."
            msg += " (Hint: try using a string which only contains numbers)"
            raise TypeError(msg) from e
        return other


class DiceIndexError(IndexError):
    """
    Exception raised for errors in the indexing of a Dice object.

    Attributes:
        dice (Dice): The Dice object where the error occurred.
    """

    def __init__(self, dice: Dice) -> None:
        """
        Initialize the DiceIndexError for the given Dice object.

        Args:
            dice (Dice): The Dice object where the error occurred.
        """
        msg = f"Invalid side: This Dice has sides numbered 1 to {len(dice)}."
        super().__init__(msg)



import re

import pytest  # noqa: F401, RUF100

from ttrpg_dice import d


def test_eq():
    d4 = d(4)
    assert d4 is not d(4)
    assert d4 == d(4)


def test_inequality():
    d4 = d(4)
    d6 = d(6)
    assert d4 != d6
    assert d4 != [None, 0.25, 0.25, 0.25, 0.25]


def test_no_sideeffects_add():
    d2 = d(2)
    adv = d2 + 1
    assert adv == d(2) + 1
    assert d2 == d(2)


def test_add_float():
    assert d(4) + 2.0 == d(4) + 2


def test_add_string():
    assert d(4) + "2" == d(4) + 2


def test_add_two():
    msg = re.escape("Cannot add 'two' and 'Dice'. (Hint: try using a string which only contains numbers)")
    with pytest.raises(TypeError, match=msg):
        d(4) + "two"


def test_add_None():
    with pytest.raises(TypeError, match="Cannot add 'NoneType' and 'Dice'"):
        d(4) + None


def test_no_sideeffects_rmul():
    d4 = d(4)
    roll = 2 * d4
    assert roll == 2 * d(4)
    assert d4 == d(4)


def test_floatxDice():
    assert 2.0000001 * d(4) == 2 * d(4)


def test_stringxDice():
    assert "2" * d(4) == 2 * d(4)


def test_twoxDice():
    msg = re.escape("Cannot multiply 'two' by 'Dice'. (Hint: try using a string which only contains numbers)")
    with pytest.raises(TypeError, match=msg):
        "two" * d(4)


def test_NonexDice():
    with pytest.raises(TypeError, match="Cannot multiply 'NoneType' by 'Dice'"):
        None * d(4)


def test_cannot_change_probabilities():
    d4 = d(4)
    msg = re.escape("You cannot change a Dice's probabilities, create a new Dice instead.")
    with pytest.raises(AttributeError, match=msg):
        d4._probabilities = [1, 2]  # noqa: SLF001


def test_individualrolls():
    die = d.from_contents({2: 1, 4: 2, 1: 3})
    assert list(die._individual_dice_rolls()) == [[1, 2], [1, 2, 3, 4], [1, 2, 3, 4], [1], [1], [1]]  # noqa: SLF001


def test_immutable():
    d4 = d(4)
    assert d4.contents[4] == 1
    with pytest.raises(TypeError, match="Dice contents cannot be changed"):
        d4.contents[4] = 2
    assert d4.contents[4] == 1


def test_no_unwanted_mutations():
    d4 = d(4)
    assert len(d4.contents) == 1
    cached_hash = hash(d4)
    assert d4.contents[5] == 0  # does not contain a d5
    assert len(d4.contents) == 1  # checking for d5 did not add an entry 5 : 0 to the contents, as a defaultdict would
    assert hash(d4) == cached_hash  # or change the hash


def test_remove_zeros():
    dice = d.from_contents({1: 0, 4: 1, 3: 0, 6: 2, 8: 0})
    assert dice.contents == {4: 1, 6: 2}



"""Rolling multiple dice."""

from __future__ import annotations

from collections.abc import Mapping
from itertools import zip_longest
from math import comb
from typing import TYPE_CHECKING

from matplotlib import pyplot as plt
from tabulate2 import tabulate

if TYPE_CHECKING:
    from collections.abc import Iterable

    from matplotlib.figure import Figure
    from mpl_toolkits.mplot3d.axes3d import Axes3D

    from .dice import Dice


def lazyroll(numdice: int, dicetype: int, target: int) -> list[int]:
    """
    Calculate equivalent single roll instead of rolling multiple dice targetting the same success value.

    Arguments:
        numdice: the number of identical dice to roll
        dicetype: the number of faces on the dice to roll
        target: the target for a successful test

    Examples:
        Instead of rolling 4d100 to see which of your (skilled) goblins hit your party with their arrows you can use:
        ```
        >>> from ttrpg_dice import lazyroll
        >>> lazyroll (4, 100, 33)
        [100, 80, 40, 11, 1]
        ```
        Then roll 1d100 and interpret the hits:
        ```
        81-100:  0 hits
         41-80:  1 hit
         12-40:  2 hits
          2-11:  3 hits
             1:  4 hits
        ```
    """
    if target > dicetype:
        msg = f"Good luck rolling {target} on a d{dicetype}!"
        raise ValueError(msg)

    def _p(hits: int) -> float:
        """Calculates the probability of an exact number of hits."""
        misses = numdice - hits
        p_successes = (target / dicetype) ** hits
        p_fails = (1 - (target / dicetype)) ** (misses)
        return p_successes * p_fails * comb(numdice, hits)

    probs = [_p(hits) for hits in range(numdice + 1)]
    return [round(sum(probs[i:]) * dicetype) for i, _ in enumerate(probs)]


class LazyRollTable:
    """Table of values for lazyrolls of varying numbers of goblins."""

    def __init__(self, maxdice: int, dicetype: int, target: int) -> None:
        """Create a table of lazyrolls for up to `maxdice`."""
        self._maxdice = maxdice
        self._maxdicerange = range(maxdice + 1)
        """use to iterate from `0` to `maxdice` rolls `for i in self._maxdicerange`"""
        self._dicetype = dicetype
        self._target = target
        self.rolls = [lazyroll(i, dicetype, target) for i in self._maxdicerange]
        """List of lists of resulting lazyrolls - (0-indexed, so _includes_ 0 dice and 0 hits)"""

    def __eq__(self, value: object) -> bool:
        """Compare self.rolls if `value` is not another `LazyRollTable`."""
        if not isinstance(value, LazyRollTable):
            return self.rolls == value
        return self.rolls == value.rolls

    def __repr__(self) -> str:  # noqa: D105
        return f"LazyRollTable for up to {self._maxdice}d{self._dicetype} targeting {self._target}: {self.rolls}"

    def __str__(self) -> str:
        """Format as a nice table ignoring zero dice and zero hits."""
        tab = "\t"
        newline = "\n"

        def _formatroll(numdice: int, rolls: list[int]) -> str:
            lazytargets = tab.join(str(lazytarget) for lazytarget in rolls[1:])
            return tab.join([str(numdice), lazytargets])

        description = f"Lazyroll table for up to {self._maxdice}d{self._dicetype} targeting {self._target} for success:"
        table_header = f"\tHITS\n\t{tab.join(str(i) for i in self._maxdicerange[1:])}"
        table_lines = [_formatroll(d, r) for d, r in enumerate(self.rolls)]
        return newline.join([description, "", table_header, *table_lines[1:]])


class PoolComparison:
    """Comparison of related dicepools."""

    def __init__(self, pools: dict[str, Dice] | Iterable[Dice], outcomes: dict[str, slice]) -> None:
        """Create comparison based on dict of named pools and dict of named outcomes."""
        if isinstance(pools, Mapping):
            self.pools = pools
        else:
            self.pools = {pool: pool for pool in pools}
        self.outcomes = outcomes
        self.chances = {
            (pool, outcome): sum(die[index])
            for pool, die in self.pools.items()
            for outcome, index in self.outcomes.items()
        }
        """Dict of chances indexed by (pool, outcome)"""

    def __str__(self) -> str:
        """Nicely formatted table."""
        data = [[pool] + [self.chances[pool, outcome] * 100 for outcome in self.outcomes] for pool in self.pools]
        headers = ["pool", *self.outcomes]
        return tabulate(data, headers=headers, tablefmt="plain", floatfmt=".2f")

    def plot(self) -> tuple[Figure, Axes3D]:
        """Plot as a 3d Bar with matplotlib and return the Axes."""
        fig, ax = plt.subplots(subplot_kw={"projection": "3d"})
        ax.bar3d(**self.plotable(), shade=True)
        ax.set_yticks([y + 0.5 for y, _ in enumerate(self.pools)], [str(pool) for pool in self.pools])
        ax.set_xticks([x + 0.5 for x, _ in enumerate(self.outcomes)], [str(outcome) for outcome in self.outcomes])
        return fig, ax

    def plotable(self) -> dict[str, list]:
        """Return bar location and sizes suitable for passing directly to matplotlib bar3d()."""
        x_locations = []
        y_locations = []
        z_locations = []
        dx_widths = []
        dy_widths = []
        dz_heights = []
        colours = []
        alphas = [
            0,  # -Z
            1,  # +Z (top)
            0,  # -Y
            0,  # +Y
            0.2,  # -X
            0.2,  # +X
        ]
        poolcolours = "bgrcmy"

        # Basic Colour Cycle (per pool)
        # ("b",0),("b",1),("b",0),("b",0),("b",0.2),("b",0.2),
        # ("g",0),("g",1),("g",0),("g",0),("g",0.2),("g",0.2),
        # ("r",0),("r",1),("r",0),("r",0),("r",0.2),("r",0.2),
        # ("c",0),("c",1),("c",0),("c",0),("c",0.2),("c",0.2),
        # ("m",0),("m",1),("m",0),("m",0),("m",0.2),("m",0.2),
        # ("y",0),("y",1),("y",0),("y",0),("y",0.2),("y",0.2),

        for x, outcome in enumerate(self.outcomes.keys()):
            for y, pool in enumerate(self.pools.keys()):
                x_locations.append(x)
                y_locations.append(y)
                z_locations.append(0)
                dx_widths.append(1)
                dy_widths.append(1)
                dz_heights.append(self.chances[(pool, outcome)])
                colours.extend(
                    zip_longest(poolcolours[y % len(poolcolours)], alphas, fillvalue=poolcolours[y % len(poolcolours)]),
                )

        return {
            "x": x_locations,
            "y": y_locations,
            "z": z_locations,
            "dx": dx_widths,
            "dy": dy_widths,
            "dz": dz_heights,
            "color": colours,
        }



"""Various functions for TTRPG Gamesmasters to help with dice rolls."""

from .dice import Dice as d  # noqa: N813
from .manydice import LazyRollTable, PoolComparison, lazyroll
from .statblock import StatBlock, statblock



import pytest  # noqa: F401, RUF100

from ttrpg_dice import LazyRollTable, lazyroll


def test_2d100_target33():
    assert lazyroll(2, 100, 33) == [100, 55, 11]


def test_5d100_target33_haszeroresult():
    assert lazyroll(5, 100, 33) == [100, 86, 53, 20, 4, 0]


def test_3d100_target41_roundingto100():
    assert lazyroll(3, 100, 41) == [100, 79, 37, 7]


def test_2d4_target2():
    assert lazyroll(2, 4, 2) == [4, 3, 1]


def test_0d100():
    assert lazyroll(0, 100, 33) == [100]


def test_guaranteedhit():
    assert lazyroll(3, 20, 20) == [20, 20, 20, 20]


def test_guaranteedfail():
    with pytest.raises(ValueError, match="Good luck rolling 13 on a d12!"):
        lazyroll(3, 12, 13)


def test_guaranteedfail_LazyTable():
    with pytest.raises(ValueError, match="Good luck rolling 13 on a d12!"):
        LazyRollTable(3, 12, 13)


def test_lazytable_eq_lazytable():
    assert LazyRollTable(4, 100, 33) == LazyRollTable(4, 100, 33)


def test_lazytable_upto4d100_target33():
    assert LazyRollTable(4, 100, 33) == [
        [100],
        [100, 33],
        [100, 55, 11],
        [100, 70, 25, 4],
        [100, 80, 40, 11, 1],
    ]


def test_lazytable_upto4d100_target33_pretty():
    output = r"""Lazyroll table for up to 4d100 targeting 33 for success:

	HITS
	1	2	3	4
1	33
2	55	11
3	70	25	4
4	80	40	11	1"""
    assert str(LazyRollTable(4, 100, 33)) == output
